from typing import Tuple

import cv2
import numpy as np

from .image_filter import image_filter
from ..helpers.remove_none_values import remove_none_values
from .kernels import validate_odd_size
from ..helpers.types import RangedNumber


def convert_to_hsv(image: np.ndarray) -> np.ndarray:
    """
    Converts an image to hsv - Mainly for beginner use
    
    :param image: image to be converted
    :return: the converted image
    """
    return cv2.cvtColor(image, cv2.COLOR_BGR2HSV)


@image_filter
def sharpen_image(image: np.ndarray, size: tuple = (3, 3)) -> np.ndarray:
    """
    Sharpens an image by preforming convolution it with a sharpening matrix

    :param image: the image (numpy array)
    :param size: the size of the sharpening matrix
    :return: the new sharpened image
    """

    validate_odd_size(size)
    kernel = np.ones(size)
    kernel *= -1
    kernel[int((size[0] - 1) / 2), int((size[1] - 1) / 2)] = kernel.size
    return cv2.filter2D(image, -1, kernel)


@image_filter
def adaptive_brightness(image: np.ndarray, brightness: RangedNumber(0, 100) = 50, hsv: bool = False) -> np.ndarray:
    """
    Changes the brightness of every pixel so that the average of the image is the target average

    :param image: The image to be changed (Numpy array)
    :param brightness: the target average for the image
    :param hsv: bool noting if the image is in hsv
    :return: a copy of the image changed
    """
    image = image if hsv else cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
    brightness = brightness * 2.55
    image_hue_mean = cv2.mean(image)[2]
    increase = brightness - image_hue_mean
    vid = image[:, :, 2]
    if increase > 0:
        vid = np.where(vid + increase <= 255, vid + increase, 255)
    else:
        vid = np.where(vid + increase >= 0, vid + increase, 0)
    image[:, :, 2] = vid
    image = image if hsv else cv2.cvtColor(image, cv2.COLOR_HSV2BGR)
    return image


@image_filter
def change_brightness(image: np.ndarray, change: float = 25, hsv_image: bool = False) -> np.ndarray:
    """
    Changes the brightness of every pixel of a BGR image by the given amount

    :param image: The image to be changed (Numpy array)
    :param change: the change (integer) (The min brightness is 0 and max is 100)

    :param hsv_image: bool noting if the image is in hsv
    :return: a copy of the image changed
    """
    image = image if hsv_image else cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
    change = change * 2.55
    vid = image[:, :, 2]
    if change > 0:
        vid = np.where(vid + change <= 255, vid + change, 255)
    else:
        vid = np.where(vid + change >= 0, vid + change, 0)
    image[:, :, 2] = vid
    image = image if hsv_image else cv2.cvtColor(image, cv2.COLOR_HSV2BGR)
    return image


def _rotate_shortcut(image, shortcut_angle):
    """
    Return a copy of the image rotated 90 degrees to the left (_counter-clockwise)

    :param image: numpy array, image to be rotated
    :return: a copy of the image rotated.
    """
    (height, width) = image.shape[:2]
    center = (width / 2, height / 2)
    rotation_matrix = cv2.getRotationMatrix2D(center, shortcut_angle, 1.0)
    return cv2.warpAffine(image, rotation_matrix, (height, width))


SHORTCUT_ANGLES = {
    90: _rotate_shortcut,
    180: _rotate_shortcut,
}


@image_filter
def rotate_image(image: np.ndarray, angle: int = 180) -> np.ndarray:
    """
    Rotates an image by a given amount of degrees.
    Note that the rotated image's dimensions will most likely change if
    the angle is not 90, -90, 180, 0 or a multiplication of them)

    :param image: the image to be rotated
    :param angle: the angle to rotate the image in degrees (positive is to the left, negative to the right)
    :return: the rotated image
    """
    angle = abs(angle % 360)
    if angle in SHORTCUT_ANGLES:
        return SHORTCUT_ANGLES[abs(angle)](image, angle)
    elif angle == 0:
        return image
    else:
        return _rotate_by_angle(image, angle)


def _rotate_by_angle(image, angle):
    """
    Rotates the given image by a given angle
    The new image will have different dimensions if the rotation angle isn't a multiple of 90

    This is an inner function that is not an image_filter.
    Use rotate_by_angle (no underscore) to rotate an image, using shortcuts for certain angles


    :param image: the image to be rotated
    :param angle: the angle of rotation
    :return: the rotated image
    """
    (height, width) = image.shape[:2]
    center_xy = (width / 2, height / 2)
    rotation_matrix = cv2.getRotationMatrix2D(center_xy, -angle, 1.0)
    cos = np.abs(rotation_matrix[0, 0])
    sin = np.abs(rotation_matrix[0, 1])
    rotated_image_dimensions = (int((height * sin) + (width * cos)), int((height * cos) + (width * sin)))
    rotation_matrix[0, 2] += (rotated_image_dimensions[0] / 2) - center_xy[0]
    rotation_matrix[1, 2] += (rotated_image_dimensions[1] / 2) - center_xy[1]
    return cv2.warpAffine(image, rotation_matrix, rotated_image_dimensions)


@image_filter
def non_local_mean_denoising(image, h=10, hColor=None, template_window_size=None, search_window_size=None,
                             destination=None):
    """
    Non local mean denoising is an image noise removal function.
    Non local mean denoising removes noise by finding matching patterns in other parts of the image
    It has different arguments for greyscale images and color images

    :param image: the image to be denoised
    :param hColor:
    :param h: parameter deciding filter strength. Higher h value removes noise better,
     but removes details of image also. (10 is default)
     h maps to h in the greyscale version of the opencv function and to hColor in the color version
    :param template_window_size: size of the template window, should be an odd number
    :param search_window_size: size of the search window
    :param destination: an image of the same size to place the result
    :return: the denoised image, a numpy array

    For more information about the implementation please refer the the opencv source code
    and this tutorial:
    https://docs.opencv.org/3.4/d5/d69/tutorial_py_non_local_means.html

    For more information on the algorithim behind the implementation look at this
    paper:
    http://www.ipol.im/pub/art/2011/bcm_nlm/
    """
    parameters = {"h": h,
                  "hColor": hColor,
                  "templateWindowSize": template_window_size,
                  "seasonWindowSize": search_window_size,
                  "dst": destination
                  }
    return cv2.fastNlMeansDenoisingColored(image,
                                           **remove_none_values(parameters))


@image_filter
def gaussian_blur(image, kernel_size=(3, 3), sigma_x=5, sigma_y=None, border_type=None):
    """
    An image filter version of cv2.gaussianBlur.

    Gaussian blur is a common filter that is used to remove image noise.

    It is considered necessary when preforming edge detection (such as using the CannyThreshold)

    Gaussian Blur assumes the noise in the image is random and that the neighbors of each pixel
    are similar in a normally distributed fashion, meaning the closer the neighbor the more it should
    be similar to the pixel. The process of applying the kernel to the image is called convolution.

    :param image: the image to be blurred (numpy array)
    :param kernel_size: the size of the window that moves over the image
    this determines what are the pixel's neighbors. The kernel must be of odd dimensions
    so that it has a center pixel - which is where the output is placed
    :param sigma_x: standard distribution of the gaussian function on the x axis,
    the larger this is the more further neighbors affect the new value of the window's center
    :param sigma_y: standard distribution of the gaussian function on the y axis,
    the larger this is the more further neighbors affect the new value of the window's center
    if this is set to 0 or None then it will take the value of sigma_x
    :param border_type: Specifies image boundaries while kernel is applied on image borders.

    More information on these can be found in the opencv's gaussianBlur function
    https://opencv-python-tutroals.readthedocs.io/en/latest/py_tutorials/py_imgproc/py_filtering/py_filtering.html#gaussian-filtering
    :return: the blurred image
    """
    parameters = {"ksize": kernel_size,
                  "sigmaX": sigma_x,
                  "sigmaY": sigma_y,
                  "borderType": border_type}
    return cv2.GaussianBlur(image, **remove_none_values(parameters))


@image_filter
def crop_image(image, point: Tuple[int, int], dimensions: Tuple[int, int]):
    """
    Crops a given rectangle from a given image, this can be used to "cut out"
    a rectangle part of the image.

    The rectangle is defined by the origin (point) a tuple of (x, y) denoting the top left corner
    and by the dimensions of its sides - defined using a tuple of (width, height).

    :param image: an image (numpy array)
    :param point: (x, y) coordinates that is the top left corner of the rectangle
    :param dimensions: (width, height) of the rectangle
    :return: the region of the image
    """
    x, y = point
    width, height = dimensions
    return image[x: x + width, y: y + height]


@image_filter
def undistort(image, camera_matrix, distortion_coefficients, destination=None, new_camera_matrix=None):
    """
    Using calculated camera matrix and distortion coefficients can be used to remove distortions caused
    by the camera and manufacturing flaws. Getting the camera matrix and distortion coefficients requires
    performing camera calibration (usually using a chessboard)
    For more information on camera calibration:
    https://opencv-python-tutroals.readthedocs.io/en/latest/py_tutorials/py_calib3d/py_calibration/py_calibration.html


    :param image: the ndarray of the image
    :param camera_matrix: the camera matrix that was calculated from the camera calibration
    :param distortion_coefficients: the distortion coefficients that were calculated from the camera calibration
    :param destination: the image the result should be saved in, None if just return
    :param new_camera_matrix: the new optimal camera matrix .
    :return: the undistorted image
    """
    parameters = {
        "destination": destination,
        "new_camera_matrix": new_camera_matrix
    }
    return cv2.undistort(image, camera_matrix, distortion_coefficients, **remove_none_values(parameters))



import numpy as np
import math


def y_intersection(line_slope, intercept, x_value):
    """
    Finds the y value of the line (y = mx + b) at point x and returns the point
    This basically solves y = mx + b

    :param line_slope: slope of the line (m)
    :param intercept: the intercept of the line (b)
    :param x_value: the value to be used (substituted with x)
    :return: the y value
    """
    return x_value, line_slope * x_value + intercept


def x_intersection(line_slope, intercept, y_value):
    """
    Calculates the x value of which the line according to the given y value
    This basically solves y = mx + b

    :param line_slope: slope of the line (m)
    :param intercept: the intercept of the line (b)
    :param y_value: the value to be used (substituted with x)
    :return: the y value
    """
    return (y_value - intercept) / float(line_slope), y_value if line_slope != 0 else (0, y_value)


def slope(first_point, second_point):
    """
    Returns the slope between 2 points

    :param first_point: first point (x,y)
    :param second_point: second point (x,y)
    :return: the slope
    """
    return 0. if first_point[0] == second_point[0] else\
        ((float(first_point[1]) - float(second_point[1])) / (float(first_point[0]) - float(second_point[0])))


def angle_between_points(first_point, second_point):
    """
    Calculates the angle between the 2 points and the horizontal axis
    :param first_point: the first point (2d)
    :param second_point: the second point  (2d)
    :return: the angle
    """
    return math.degrees(math.atan(slope(first_point, second_point)))


def contour_angle_line(contour):
    """
    Returns the angle of the line created by the top-most and
    bottom-most points of a contour

    :param contour: the contour, a numpy array
    :return: the angle in degrees
    """
    if type(contour) != np.ndarray:
        raise TypeError("Contour must be a numpy array")
    topmost = tuple(contour[contour[:, :, 1].argmin()][0])
    bottommost = tuple(contour[contour[:, :, 1].argmax()][0])
    return angle_between_points(topmost, bottommost)


def circle_area(radius):
    """
    Calculates the area of a circle with the given radius.

    :param radius: The radius of the circle
    :return: Area of the circle
    """
    return math.pi * (radius ** 2)


def distance_between_points(first_point, second_point):
    """
    Calculates the Distance between 2 points. (x^2 + y^2) ^ 0.5

    :param first_point: tuple of x and y value of point 1
    :param second_point: tuple of x and y value of point 2
    :return: Float value of the distance between the 2 points
    :rtype: float
    """
    return ((first_point[0] - second_point[0]) ** 2 + (first_point[1] - second_point[1]) ** 2) ** 0.5


def regular_polygon_angle(side_amount: int) -> float:
    """
    Returns the inner angle of a polygon with the given amount of sides

    :param side_amount: amount of sides the polygon has
    :return:
    """
    return (side_amount - 2) * (180 / side_amount)


def polygon_area(side_length, amount_of_sides):
    """
     area of a polygon of
    :param side_length: side_length of a side of the polygon
    :param amount_of_sides: Amount of sides the polygon has
    :return: the area of the polygon
    """
    angle = 180 / amount_of_sides
    return 0.25 * amount_of_sides * (side_length ** 2) * (math.cos(math.radians(angle)) / math.sin(math.radians(angle)))


def law_of_cosine(first_point, second_point, third_point) -> float:
    """
    Finds the angle BAC  in degrees given that first_point = A second point = B and third point = C
    """
    first_length = distance_between_points(first_point, second_point)
    second_length = distance_between_points(first_point, third_point)
    third_length = distance_between_points(second_point, third_point)
    angle = (first_length ** 2 + second_length ** 2 - third_length ** 2) / (first_length * second_length * 2)
    return math.degrees(angle)





from abc import abstractmethod


class Threshold:
    """
    Threshold is a base class for Threshold object that threshold an image
    (Color, binary or otherwise), threshold is then followed up
    """
    @abstractmethod
    def convert(self, image):
        pass

    @abstractmethod
    def validate(self, *args, **kwargs) -> bool:
        pass



import math
import typing
from typing import Union

import numpy as np

from ..math.contours import contour_center


def horizon_angle(point: Union[np.ndarray, typing.Tuple[int, int]], field_of_view: float, image_width: int) -> float:
    """
    Returns the angle compared to the center of the image for a given field of view image width and point

    :param point: the contour or its center
    :param field_of_view: the horizontal field of view
    :param image_width: the image width
    :return: the angle compared to the center of the image , negative left positive right.
    """
    point = point[0] if type(point) in (tuple, list, set) else point
    x_value = contour_center(point)[0] if type(point) is np.ndarray else point
    angle = math.atan((x_value - (image_width - 1) / 2) / float(focal_length(image_width, field_of_view)))
    return float(math.degrees(angle))


def vertical_angle(point: Union[np.ndarray, typing.Tuple[int, int]], field_of_view: float, image_height: int) -> float:
    """
    Returns the angle compared to the center of the image for a given field of view image height and point

    :param point: the x val or center
    :param field_of_view: the vertical field of view
    :param image_height: the image height
    :return: the angle compared to the center of the image , negative up positive down.
    """
    point = point[1] if type(point) in (tuple, list, set) else point
    y_val = contour_center(point)[1] if type(point) is np.ndarray else point
    angle = math.atan((y_val - (image_height - 1) / 2) / float(focal_length(image_height, field_of_view)))
    return float(math.degrees(angle))


def focal_length(image_width, field_of_view):
    """
    Calculates the focal length in pixels of the camera for a given an image width and field of view

    :param image_width: width of the image in pixels
    :param field_of_view:
    :return: calculates the focal length, a float
    """
    return image_width / float((2 * math.tan(math.radians(float(field_of_view / 2)))))



import cv2
import numpy as np

from .detector import Detector


class HaarCascadeDetector(Detector):
    """
    A detector used to detect objects using haar cascade algorithim
    The Detector initializes using an xml file containing the descriptor

    The Detector uses the underlying cv2.CascadeClassifier

    For more information on

    """
    def __init__(self, classifier: str):
        self.classifier_source = classifier
        self.classifier = cv2.CascadeClassifier(classifier)

    def detect(self, image: np.ndarray, *args, **kwargs):
        greyscale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        return self.classifier.detectMultiScale(greyscale)



from functools import partial

from ..math import contours
from ..math import geometry


def contour_and_point_distance(point):
    """
    A reusable function that preloads a point and then
    calculates the distance to a given contour.
    :param point: an (x,y) tuple point
    :return:
    """
    return partial(geometry.distance_between_points, point)


def contour_center_and_point_distance(point):
    def contour_center_distance(contour, second_point):
        return geometry.distance_between_points(contours.contour_center(contour), second_point)
    return partial(contour_center_distance, point)



import typing
from functools import reduce
from typing import Tuple

import cv2
import numpy as np

from .geometry import distance_between_points, law_of_cosine
from .shape_fill_ratios import circle_fill_ratio
from ..image_filters.image_filters import crop_image


def target_size(contours: typing.List[np.ndarray]) -> float:
    """
    Returns the sum of contour areas of a list of contours
    """
    return sum(list(map(cv2.contourArea, contours)))


def open_arc_length(contour: np.ndarray) -> float:
    """
    Returns the arc length of an "open" contour
    """
    return cv2.arcLength(contour, False)


def open_contour_approximation(contour: np.ndarray, approximation_coefficient: float = 0.02):
    """
    Returns the vertices of the approximation of the contour
    NOTE: for "open" (unclosed shape) contours only!

    :param contour: open contour (numpy array)
    :param approximation_coefficient: approximation coefficient
    :return: list of vertices
    """
    arc = approximation_coefficient * cv2.arcLength(contour, False)
    return cv2.approxPolyDP(contour, arc, False)


def contour_center(contour: np.ndarray) -> Tuple[float, float]:
    """
    Returns the coordinates center of a contour

    :param contour: a single contour
    :return: X coordinate of center, Y coordinate of center
    """
    moments = cv2.moments(contour)
    try:
        area = float(moments['m00'])
        center_x, center_y = moments['m10'] / area, moments['m01'] / area
    except ZeroDivisionError:
        raise ValueError('Contour given is too small!,'
                         'try using area_filter to remove small contours (try min_area=20)')
    return center_x, center_y


def _contour_center_sum(point_sum, contour):
    """
    Helper function for contour_average_center
    """

    def _average_point_reduce(first_point, second_point):
        return first_point[0] + second_point[0], first_point[1] + second_point[1]

    current_contour_center = contour_center(contour)
    return _average_point_reduce(point_sum, current_contour_center)


def contour_average_center(contours) -> Tuple[float, float]:
    """
    Calculates the average center of a list of contours

    :param contours: the list of contours
    :return: the average center  (x,y)
    """
    contour_amount = len(contours)
    point_sum = reduce(_contour_center_sum, contours, (0, 0))
    average = (point_sum[0] / contour_amount)
    return average


def contour_approximation(contour, approximation_coefficient=0.02):
    """
    Returns the approximation of the contour

    :param contour: a numpy array
    :param approximation_coefficient: the coefficient of the contour length approximation
    :return:
    """
    perimeter = cv2.arcLength(contour, True)
    approximation = cv2.approxPolyDP(contour, approximation_coefficient * perimeter, True)
    return approximation


def contour_lengths_and_angles(contour, approximation_coefficient=0.02):
    """
    Returns a list of lengths and angles of a given contour.

    :param contour: contour (numpy array) to find its
    :param approximation_coefficient:
    :return: returns the list of sides
    """
    approximation = contour_approximation(contour, approximation_coefficient)
    vertices = []
    lengths = []
    angles = []
    for points in approximation:
        vertex = points[0][1], points[0][0]
        vertices.append(vertex)

    for index, point in enumerate(vertices):
        vertex_index = index + 1 if index != len(approximation) - 1 else 0
        current_length = distance_between_points(point, vertices[vertex_index])
        current_angle = law_of_cosine(vertices[index - 1], point, vertices[vertex_index])
        lengths.append(current_length)
        angles.append(current_angle)

    return vertices, lengths, angles


def calculate_normalized_screen_space(contours: typing.Union[typing.List[np.ndarray], np.ndarray], image):
    """


    :param contours:
    :param image:
    :return:
    """
    height, width, _ = image.shape
    x, y = contour_average_center(contours)
    x /= width / 2
    y /= height / 2
    return x - 1, y - 1


def circle_rating(contour, area_factor=0.9, radius_factor=0.8):
    """
    Returns a rating of how close is the circle to being a circle

    :param contour: the contour that its rating is calculated
    :param area_factor: the factor (p-value) that separates an area ratio that is a circle and one that isn't
    :param radius_factor: the factor (p-value) that separates an radius ratio that is a circle and one that isn't
    :return: the circle rating for the given contour
    """
    fill_ratio, radius = circle_fill_ratio(contour)
    _, _, width, height = cv2.boundingRect(contour)
    radius_ratio = ((((radius * 2) ** 2) / (float(width) * height)) ** 0.5)
    rating = (radius_ratio * radius_factor) * (fill_ratio * area_factor)
    return rating


def crop_contour_region(contour: np.ndarray, image: np.ndarray):
    """
    Returns a region of the image where the contour (using its bounding box)

    :param contour: the object found in the image, a numpy array
    :param image: the image from which the contour was found
    :return: the image  region that contains the contour
    """
    corner_x, corner_y, width, height = cv2.boundingRect(contour)
    return crop_image(image, (corner_x, corner_y), (width, height))


def contour_average_color(contours: typing.List[np.ndarray], image: np.ndarray):
    """
    Calculates the average color (in hsv) of all the pixels of a list of contours in the image.

    :param contours: a list of contour(s) detected in the image
    :param image: the image where the contours where detected, numpy array
    :return: hsv color (h, s, v) tuple of the average color
    """
    hsv = cv2.cvtColor(image, cv2.COLOR_RGB2HSV)
    mask = np.zeros(image.shape, np.uint8)
    if isinstance(contours, np.ndarray):
        contours = [contours]
    cv2.drawContours(mask, contours, -1, 255, -1)
    return cv2.mean(hsv, mask)



from functools import reduce
from typing import List

import cv2
import numpy as np

from .detector import Detector
from ..thresholds.threshold import Threshold
from ..partials.filter_applier import filter_applier


class ThresholdDetector(Detector):
    """
    ThresholdDetector is a detector used to find contours in a binary image.
    The binary image is created using a Threshold object.
    Examples of this are binary threshold, color, multicolor thresholds.


    For more information on binary thresholding and color thresholding refer to
    the documentation of the relevant Threshold object and:
    Binary Thresholding - https://docs.opencv.org/3.4/d7/d4d/tutorial_py_thresholding.html
    Color Thresholding - https://docs.opencv.org/master/da/d97/tutorial_threshold_inRange.html


    Threshold also allows the usage of morphological functions
    which are functions that are applied on the binary image created by the threshold.

    .. note::

        Morphological functions are functions decorated with @image_filter that act on binary images.

    Examples for morphological functions are erosion or dilation.


    For more information on morphological functions:
    https://docs.opencv.org/3.4/d9/d61/tutorial_py_morphological_ops.html
    """

    def __init__(self, threshold: Threshold, morphological_functions):
        """
        :param threshold: a Threshold object used to create binary images
        :param morphological_functions: a list of morphological functions
        """
        self.morphological_functions = morphological_functions
        self.threshold = threshold

    def apply_threshold(self, image: np.ndarray, threshold=None) -> np.ndarray:
        """
        Gets a mask (binary image) for a given image and Threshold object
        (uses self.threshold if given threshold was none)

        :param image: the numpy array of the image
        :param threshold: the Threshold object used to create the binary mask
        :return: the binary mask

        """
        threshold = threshold or self.threshold
        return threshold.convert(image)

    def find_contours_in_mask(self, mask: np.ndarray, return_hierarchy=False, apply_morphs=True
                              ) -> List[np.ndarray]:
        """
        This function is used to find and extract contours (object shapes) from a binary image
        (image passed through a threshold)

        :param mask: binary image (mask), a numpy array
        :param return_hierarchy: if the hierarchy should be returned
        :param apply_morphs: if the morphological functions should be applied.
        :return: the list of contours
        """
        mask = self.apply_morphological_functions(mask) if apply_morphs else mask
        result = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        if len(result) == 3:
            _, contours, hierarchy = result
        elif len(result) == 2:
            contours, hierarchy = result
        else:
            raise ValueError("Invalid output from cv2.findContours, check that your cv2 (OpenCV) version is supported")
        return (contours, hierarchy) if return_hierarchy else contours

    def detect(self, image: np.ndarray, return_hierarchy=False, *args, **kwargs) -> List[np.ndarray]:
        """
        Gets a list of all the contours within the threshold that was given

        :param image: image from which to get the contours
        :param return_hierarchy: if the hierarchy should be returned
        :return: list of all contours matching the range of hsv colours

        """
        image_mask = self.apply_threshold(image)
        return self.find_contours_in_mask(image_mask, return_hierarchy=return_hierarchy)

    def apply_morphological_functions(self, mask, morphological_functions=None):
        """
        Applies all morphological functions on the mask (binary images) created using the threshold,
        Morphological functions are functions that are applied
        to binary images to alter the shape of "detected" regions


        :param mask: the mask on which the functions should be applied
        :param morphological_functions: list of morphological_functions to be
         applied instead of self.morphological_functions
        :return: the applied mask
        """
        if type(self.morphological_functions) not in (tuple, list, set):
            return mask
        morphological_functions = morphological_functions or self.morphological_functions
        return reduce(filter_applier, morphological_functions, mask)



from typing import List, Any


class Detector:
    """
    Detector the class that implements the Detection stage of the pipeline

    The only required function is detect which should get an image and return

    """
    def detect(self, image, *args, **kwargs) -> List[Any]:
        raise NotImplemented()
